import csv
import os
import glob

from typing import List

from .utils.types import *
import pkg_resources

COLUMNS = ("Key", "English", "French", "German", "Polish", "Spanish", "Italian",
            "Swedish", "Czech", "Hungarian", "Dutch", "Portuguese", "Russian", "Finnish")


class GameFile:
    def __init__(self, localisation_folder=pkg_resources.resource_listdir("pyvic2waranalyzer", "localisation"), lang="English"):
        """Initializes the parser

        Parameters
        ----------
        localisation_folder: :class:`str` or :class:`list` or :class:`None`
            Indicates a localisation folder with .csv files.
        lang: :class:`str`
            Indicates the language to translate.
        """
        self.localisations = {}
        self.file = None
        self.reader = None
        self.lang_index = 1
        for pos, language in enumerate(COLUMNS, 0):
            if language == "Key":
                continue
            if lang.lower() == language.lower():
                self.lang_index = pos
        # print(localisation_folder)
        if localisation_folder:
            if isinstance(localisation_folder, list):
                for a in localisation_folder:
                    filename = pkg_resources.resource_filename("pyvic2waranalyzer", os.path.join("localisation", a))
                    # print(filename)
                    # self.file = pkg_resources.resource_string("pyvic2waranalyzer", os.path.join("localisation", a)).decode("latin-1")
                    if filename.endswith(".csv"):
                        with open(filename, "r", newline="", encoding="latin-1", errors="ignore") as self.file:
                            self.file.seek(0)
                            self.reader = csv.reader(self.file, delimiter=";")
                            next(self.reader)
                            for _ in self.reader:
                                try:
                                    # print(_)
                                    self.localisations.update({_[0]: _[self.lang_index] if _[self.lang_index] != "" else _[1]})
                                except IndexError:
                                    pass
            else:
                for filename in glob.glob(os.path.join(localisation_folder, "*.csv")):
                    with open(filename, "r", newline="", encoding="latin-1", errors="ignore") as self.file:
                        self.file.seek(0)
                        self.reader = csv.reader(self.file, delimiter=";")
                        next(self.reader)
                        for _ in self.reader:
                            try:
                                self.localisations.update({_[0]: _[self.lang_index] if _[self.lang_index] != "" else _[1]})
                            except IndexError:
                                pass
        self.sl = None
        self.attackerDefender = None
        self.battleProcessing = False
        self.warGoalProcessing = False
        self.OriginalwarGoalProcessing = False
        self.warProcessing = False
        self.war = []
        self.iter_ = 0
        self.wargoalcounter = 0
        self.warcounter = 0
        self.bracketCounter = 0
        self.wargoal_disabled = False

    def scan(self, filename):
        """Scan the save file, returns a list of :class:`War`

        Parameters
        ----------
        filename: :class:`str` or :class:`bytes`

        Returns
        -------
        List[:class:`War`]
        """
        self.sl = None
        self.attackerDefender = None
        self.battleProcessing = False
        self.warGoalProcessing = False
        self.OriginalwarGoalProcessing = False
        self.warProcessing = False
        self.war = []
        self.iter_ = 0
        self.wargoalcounter = 0
        self.warcounter = 0
        self.bracketCounter = 0
        self.wargoal_disabled = False
        if isinstance(filename, bytes):
            doc = filename.decode("latin-1", errors="ignore")
        elif isinstance(filename, str):
            with open(filename, "r", errors="ignore", encoding="latin-1") as f:
                doc = f.read()
        self.sl = doc.split("\n")
        for i in range(len(self.sl) - 3):
            if self.is_previous_war(i):
                self.warProcessing = True
                # print("activado")

            if self.warProcessing:
                # print(self.sl[i])
                self.bracketCounterChange(self.sl[i])

                if "battle=" in self.sl[i] or self.battleProcessing:
                    # print(self.sl[i], "battle")
                    # print("BATALLA", self.sl[i])
                    if self.iter_ == 0 or "battle=" in self.sl[i] and not self.battleProcessing:
                        # print("created battle", self.sl[i])
                        if not self.battleProcessing:
                            self.war[self.warcounter].battles.append(Battle())

                    self.battleProcessing = True
                    self.BattleReader(self.sl[i])
                elif "war_goal=" in self.sl[i] or self.warGoalProcessing and not self.wargoal_disabled:
                    if self.wargoalcounter == 0 or "war_goal=" in self.sl[i]:
                        if not self.warGoalProcessing:
                            self.war[self.warcounter].wargoals.append(Wargoal())
                    self.warGoalProcessing = True
                    self.wargoalreader(self.sl[i])
                elif "original_wargoal=" in self.sl[i] or self.OriginalwarGoalProcessing and not self.wargoal_disabled:
                    if self.wargoalcounter == 0 or "original_wargoal=" in self.sl[i]:
                        if not self.OriginalwarGoalProcessing:
                            self.war[self.warcounter].wargoals.append(OriginalWargoal())
                    self.OriginalwarGoalProcessing = True
                    self.wargoalreader(self.sl[i])
                else:
                    # print("guerra", self.sl[i])
                    # print(self.sl[i], "war_parse")
                    self.war_parse(self.sl[i])
            try:
                if self.bracketCounter == 0 and self.warProcessing and not self.war[self.warcounter].name == "":
                    # print(self.sl[i], "next_war")
                    self.warGoalProcessing = False
                    self.wargoalcounter = 0
                    self.iter_ = 0
                    self.warcounter += 1
                    self.warProcessing = False
            except:
                pass
        if self.file:
            self.file.close()
        return self.war

    def localize(self, key):
        return self.localisations.get(key) or key

    def bracketCounterChange(self, line):
        if "{" in line:
            self.bracketCounter += 1
        elif "}" in line:
            self.bracketCounter -= 1

    def nameextractor(self, line):
        line = line.strip()
        line = line.split("=")[1].replace("'", "")
        return line.replace('"', '')

    def wargoalreader(self, line):
        # TODO: list of wargoals?
        # wargoal
        if "state_province_id=" in line:
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].state = line
        elif "casus_belli=" in line:
            # print(line)
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].casus_belli = self.localize(line)
        elif "country=" in line:
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].country = self.localize(line)
        elif "actor=" in line:
            line = self.nameextractor(line)
            # print(self.wargoalcounter, self.war[self.warcounter].wargoals[self.wargoalcounter])
            self.war[self.warcounter].wargoals[self.wargoalcounter].actor = self.localize(line)
        elif "receiver=" in line:
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].receiver = self.localize(line)
        elif "score=" in line:
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].score = float(line)
        elif "change=" in line:
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].change = float(line)
        elif "date=" in line:
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].date = line
        elif "is_fulfilled=" in line:
            # print(line)
            line = self.nameextractor(line)
            self.war[self.warcounter].wargoals[self.wargoalcounter].is_fulfilled = True if line == "yes" else False
        elif "}" in line:
            # print(line)
            if not self.war[self.warcounter].wargoals[self.wargoalcounter]:
                del self.war[self.warcounter].wargoals[self.wargoalcounter]
                self.wargoalcounter -= 1
            self.warGoalProcessing = False
            self.OriginalwarGoalProcessing = False
            self.wargoalcounter += 1

    def BattleReader(self, line):
        if "name=" in line:
            name = self.nameextractor(line)
            self.war[self.warcounter].battles[self.iter_].name = name
            self.attackerDefender = True
        elif "location=" in line:
            location = int(self.nameextractor(line))
            self.war[self.warcounter].battles[self.iter_].location = location
        elif "result=" in line:
            result_ = True if "yes" in self.nameextractor(line) else False
            self.war[self.warcounter].battles[self.iter_].result = result_
        elif "country=" in line:
            # print(line)
            line = self.nameextractor(line)
            if self.attackerDefender:
                country_atck = line
                self.war[self.warcounter].battles[self.iter_].attacker = self.localize(country_atck)
            else:
                country_def = line
                self.war[self.warcounter].battles[self.iter_].defender = self.localize(country_def)
        elif "leader=" in line:
            line = self.nameextractor(line)
            if self.attackerDefender:
                leader_atck = line
                self.war[self.warcounter].battles[self.iter_].attackerLeader = leader_atck
            else:
                leader_def = line
                self.war[self.warcounter].battles[self.iter_].defenderLeader = leader_def
        elif "losses=" in line:
            line = int(self.nameextractor(line))
            if self.attackerDefender:
                self.attackerDefender = False
                losses_atck = line
                self.war[self.warcounter].battles[self.iter_].attackerLosses = losses_atck
            else:
                losses_def = line
                self.war[self.warcounter].battles[self.iter_].defenderLosses = losses_def
                if not self.war[self.warcounter].battles[self.iter_]:
                    del self.war[self.warcounter].battles[self.iter_]
                    self.iter_ -= 1
                self.iter_ += 1
                self.battleProcessing = False

        elif "attacker=" not in line and "defender=" not in line and "{" not in line and "}" not in line and "battle=" not in line:
            if self.attackerDefender:
                line = line.strip()
                first, second = line.split("=")
                second = int(second.replace('"', ''))
                first = self.localize(first)
                self.war[self.warcounter].battles[self.iter_].attackerArmy.update({first: second})
            else:
                line = line.strip()
                first, second = line.split("=")
                second = int(second.replace('"', ''))
                first = self.localize(first)
                self.war[self.warcounter].battles[self.iter_].defenderArmy.update({first: second})

    def war_parse(self, line):
        if "name=" in line:
            line = self.nameextractor(line)
            self.war.append(War(name=line))
        elif "1=" in line:
            pass
        # elif "add_attacker=" in line:
        #     line = self.localize(self.nameextractor(line).strip())
        #     self.war[self.warcounter].attackers.append(line)
        # elif "add_defender=" in line:
        #     line = self.localize(self.nameextractor(line).strip())
        #     self.war[self.warcounter].defenders.append(line)
        elif "attacker=" in line:
            line = self.localize(self.nameextractor(line).strip())
            if line != "---":
                if line not in self.war[self.warcounter].attackers:
                    self.war[self.warcounter].attackers.insert(0, line)
            else:
                pass
        elif "defender=" in line:
            line = self.localize(self.nameextractor(line).strip())
            if line != "---":
                if line not in self.war[self.warcounter].defenders:
                    self.war[self.warcounter].defenders.insert(0, line)
            else:
                pass
        elif "action=" in line:
            line = self.nameextractor(line)
            self.war[self.warcounter].action = line

    def is_previous_war(self, index):
        return "previous_war=" in self.sl[index] or "active_war=" in self.sl[index] and "{" in self.sl[index + 1] \
            and "name=" in self.sl[index + 2] \
            and 'American War of Independence' not in self.sl[index + 2] \
            and 'Texan War of Independence' not in self.sl[index + 2] \
            and 'Ottoman Restoration of Tripoli' not in self.sl[index + 2]


from setuptools import setup, find_packages

extras_require = {
    "docs": ["sphinx==4.1.1", "sphinx-rtd-theme==0.5.2", "sphinxcontrib-napoleon", "sphinxcontrib-serializinghtml",
             "sphinxcontrib-htmlhelp", "sphinxcontrib-applehelp", "sphinxcontrib-jsmath", "myst_parser"]
}

setup(
    name='pyvic2waranalyzer',
    version='1.4',
    packages=find_packages(),
    url='https://github.com/InternetExplorer404/PyVic2WarAnalyzer',
    download_url="https://github.com/InternetExplorer404/PyVic2WarAnalyzer/archive/refs/tags/v1.4.tar.gz",
    include_package_data=True,
    license='Unlicense',
    author='Alvaro',
    author_email='',
    extra_require=extras_require,
    description='Victoria 2 War Analyzer written in Python!',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Other/Nonlisted Topic',
        'License :: OSI Approved :: The Unlicense (Unlicense)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



class Wargoal:
    """Represents a wargoal in a :class:`War`

    Attributes
    ----------
    state: :class:`int`
        A state id from 'take state' CB.
    casus_belli: :class:`str`
        The name of the casus belli.
    actor: :class:`str`
        The country which made this wargoal.
    country: :class:`str`
        TAG or country which
    receiver: :class:`str`
        The country which received this wargoal.
    date: :class:`str`
        Wargoal date.
    is_fulfilled: :class:`bool`
        May not be present, only if it's an active war.
    score: :class:`float`
        May not be present, only if it's an active war.
    change: :class:`float`
        May not be present, only if it's an active war.
    """

    aliases = {
        'casus': 'casus_belli',
        'casusbelli': 'casus_belli',
    }

    def __init__(self, state: int = None, casus_belli: str = None, actor: str = None,
                 receiver: str = None, date: str = None, is_fulfilled: bool = None,
                 score: float = None, change: float = None, country: str = None):

        self.state = state
        self.casus_belli = casus_belli
        self.country = country
        self.actor = actor
        self.receiver = receiver
        self.date = date
        self.is_fulfilled = is_fulfilled
        self.score = score
        self.change = change

    def __setattr__(self, name, value):
        name = self.aliases.get(name, name)
        object.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == "aliases":
            raise AttributeError  # http://nedbatchelder.com/blog/201010/surprising_getattr_recursion.html
        name = self.aliases.get(name, name)
        return object.__getattribute__(self, name)

    def __bool__(self):
        if self.casus_belli and self.actor and self.receiver:
            return True
        else:
            return False

    def __str__(self):
        return self.casus_belli

class OriginalWargoal(Wargoal):
    """Represents an original wargoal in a :class:`War`

    Attributes
    ----------
    state: :class:`int`
        A state id from 'take state' CB.
    casus_belli: :class:`str`
        The name of the casus belli.
    actor: :class:`str`
        The country which made this wargoal.
    receiver: :class:`str`
        The country which received this wargoal.
    """
    def __init__(self, state: int = None, casus_belli: str = None, actor: str = None, receiver: str = None):
        self.state = state
        self.casus_belli = casus_belli
        self.actor = actor
        self.receiver = receiver
        super().__init__(state=self.state, casus_belli=self.casus_belli, actor=self.actor, receiver=self.receiver)

class War:
    """Represents a war between 2 nations or more

    Attributes
    ----------
    attackers: :class:`list`
        All attackers.
    defenders: :class:`list`
        All defenders.
    battles: List[:class:`Battle`]
        A list of all battles in that war.
    wargoals: :class:`Wargoal`
        The list of wargoals.
    action: :class:`str`
    """
    def __init__(self, name: str = None,
                 action: str = None):
        self.name = name
        self.attackers = []
        self.defenders = []
        self.battles = []
        self.wargoals = []
        self.action = action

    @property
    def total_losses(self) -> int:
        """
        All losses from a :class:`War`
        """
        total = 0
        for a in self.battles:
            total += a.total_losses
        return total

    @property
    def total_army(self) -> int:
        """
        All forces from a :class:`War`
        """
        total = 0
        for c in self.battles:
            total += c.total_army
        return total

    def __str__(self):
        return self.name

    def __add__(self, other):
        if isinstance(other, War):
            return self.total_losses + other.total_losses
        else:
            return self.total_losses + other

    def __radd__(self, other):
        if other == 0:
            return self.total_losses
        else:
            self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, War):
            return self.total_losses * other.total_losses
        else:
            return self.total_losses * other

    def __rmul__(self, other):
        self.__mul__(other)

    def __bool__(self):
        if self.name and self.attackers and self.defenders:
            return True
        else:
            return False


class Battle:
    """Represents a battle in a :class:`War`

    Attributes
    ----------
    name: :class:`str`
        Indicates the battle name, aka the province name
    location: :class:`int`
        Indicates a province id
    result: :class:`bool`
        Indicates if the attacker won the battle or else
    defender: :class:`str`
    attacker: :class:`str`
    attackerLosses: :class:`int`
    defenderLosses: :class:`int`
    attackerLeader: :class:`str`
    defenderLeader: :class:`str`
    attackerArmy: :class:`dict`
    defenderArmy: :class:`dict`
    """
    def __init__(self, name: str = None, location: int = None, result: bool = None, defender: str = None,
                 attacker: str = None, attackerLosses: int = 0,
                 defenderLosses: int = 0, attackerLeader: str = None, defenderLeader: str = None):
        self.name = name
        self.location = location
        self.result = result
        self.defender = defender
        self.attacker = attacker
        self.attackerLosses = attackerLosses
        self.defenderLosses = defenderLosses
        self.attackerLeader = attackerLeader
        self.defenderLeader = defenderLeader
        self.attackerArmy = {}
        self.defenderArmy = {}

    @property
    def total_losses(self) -> int:
        """
        All losses from a :class:`Battle`
        """
        return self.defenderLosses + self.attackerLosses

    @property
    def total_army(self) -> int:
        """
        All forces from a :class:`Battle`
        """
        total = 0
        for a in self.attackerArmy.values():
            total += a
        for b in self.defenderArmy.values():
            total += b
        return total

    def __str__(self):
        return self.name

    def __add__(self, other):
        if isinstance(other, Battle):
            return self.total_losses + other.total_losses
        else:
            return self.total_losses + other

    def __radd__(self, other):
        if other == 0:
            return self.total_losses
        else:
            self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, Battle):
            return self.total_losses * other.total_losses
        else:
            return self.total_losses * other

    def __rmul__(self, other):
        self.__mul__(other)

    def __bool__(self):
        if self.name and self.defender and self.attacker:
            return True
        else:
            return False

